Module MPC_Config
======================

SUMMARY:
--------

Here, MPC_Config handles MPC XML configuration file support, mapping user needs
with library capabilities. For a complete documentation, pleasee read the
following files: userman.\* and devman.\*. This documentation will partly resume
these files.

CONTENTS:
---------
* **editor/**     : Web-based Configuratione editor (JS)
* **src/**        : Config sources files, used for interface w/ MPC source code
* **generators/** : XML Parser for each config-meta.xml files (XSL files)
* **generated/**  : Current built configuration (not written, generated by a tool)
* **doc/**        : proper MPC_Config module documentation (pdfs)
* **bin/**        : Source files for mpc_print_config executable
* **extra/**      : Extra resources : Sample config files,etc...
    

COMPONENTS:
-----------

### Introduction

Since MPC 2.4.1, a configuration system has been introduced through the module
`MPC_Config`: it enables the user to setup some parameters at the runtime when
running his binary with MPC. Some variables can be configured during the
runtime. To do this, a configuration file defines all the variables and their
values, and MPC will load it at the execution.

An example of a configuration file can be found in the
`$PREFIX/share/mpc` folder, where `$PREFIX` matches to the install
directory of MPC.

The configuration file is written in XML, and two parts make up it:
1.  A first one listing all the available profiles and their associated
    parameters;
2.  An other one defining which profiles to apply at runtime depending
    on the values of environment variables.


### config-meta.xml

Each module to be added in the configuration system needs its own
`config-meta.xml` file. The general structure of such a file is designed
as follow:

    <config>
      <usertypes>...</usertypes>
      <modules>...</modules>
    <profile>

The developer has to defined all the types (`structure`, `union`, etc.)
that will be used into the configuration system in the `usertypes`
section.

#### The `usertypes` section

In this section, the developer will define structures (`struct`), unions
(`union`) and enumerators (`enum`) that will be used in the
configuration system. Each one of them have two required attributes:
`name` (resp. `doc`) is the name (resp. the description) of the
variable/object.

A structure (as a C/C++ struct) is a set of `param` or `array` which
have required attributes:

1.  `name` which is the name of the variable;

2.  `type` which is its type (simple: `int`, `float`, …or complex:
    `struct`, `union`, …),

3.  `doc` which shortly describes what it is for,

4.  `[`only for `array``]` `entry-name` which matches to the tag name of
    the array elements.

The `param` have an optional attribute, `default`, to initialize the
variable to a specific default value.

    <struct name="my_struct" doc="This a struct test">
      <param name="param1" type="string" doc="String param." />
      <param name="param2" type="user_type" doc="Param of type user_type (enum, struct, ...)." />
      <array name="array1" entry-name="elt" type="int" doc="Array of int with elt as xml tag name." />
    </struct>

An union (as a C/C++ union) matches to a list of choices defined by a
name and a type.

    <union name="my_union" doc="This an union test">
      <!-- Define a choice with a string param -->
      <choice name="choice1" type="string" />
      <!-- Define a choice with an user type param -->
      <choice name="choice2" type="user_type" />
      <!-- Define a choice with an int param -->
      <choice name="choice3" type="int" />
    </union>

An enumerator (as a C/C++ enum) is a set of possible values as shown in
the following code.

    <enum name="day" doc="Week days">
      <value>Monday</value>
      <value>Tuesday</value>
      <value>Wednesday</value>
    </enum>

#### The `modules` section

In this section, the developer will declare the high-level structures he
wants the user to configure. In the following code, there will be two
configurable structures of type `my_struct_1` and `my_struct_2`.

    <modules>
      <module name="struct_1" type="my_struct_1" />
      <module name="struct_2" type="my_struct_2" />
    </modules>


### Build the configuration C files
A developper who wants to integrate his MPC module into the
configuration system must:

1.  Create a configuration file `config-meta.xml` in his module and
    define all the options he wants to parametrized;

2.  Mark the dependency to the MPC\_Config module by adding in the file
    `module_dep`: `need_module MPC_Config`;

3.  Regenerate the MPC\_Config auto-generated files by execution
    `./MPC_Tools/mpc_gen_runtime_config` from `mpc` directory;

4.  Include the header `sctk_runtime_config.h` in his source files;

5.  Use the function `sctk_runtime_config_get()` to access to the
    configuration structure.


### Using the configuration

A configuration file must have a default profile:
it initializes all the variables of the configuration system with their
default values.

    <profile>
      <name>default</name>
      <modules>
        <launcher>
          <nb_node>4</nb_node>
          <nb_processor>4</nb_processor>
          <share_node>false</share_node>
        </launcher>
        <allocator>
          <numa>true</numa>
          <debug>false</debug>
        </allocator>
      </modules>
    <profile>

This simple example defines three variables for a module called
`launcher`, and two for `allocator`.

The parameters inside a module can be of different types (matching to C
types) :

-   Numbers such as integers, doubles, etc.;

-   String;

-   Size (i.e. `50 MB`, `10 PB`, etc.);

-   Boolean (`true` or `false`);

-   Enum;

-   Function Pointer.

All the parametrizable variables can be listed by executing the command
`man mpc_config`: it gives information (type, default value,
description) for each variable.
